//_______ DATAMEMBERS _______

//View matrices
float4x4 mWorldViewProj : WorldViewProjection;
float4x4 mWorldView : WorldView;
float4x4 mWorld : World;
float4x4 mView : View;
float4x4 mProj : Projection;
float4x4 mViewProjInverse : ViewProjectionInverse;
float4x4 mViewProjPrevious;

//Light matrices
float4x4 mLight;
float4x4 mProjLight;

//Vectors
float3 lightColorAmbient;
float3 lightColorDiffuse;
float3 lightDirection;
const float3 viewDirection = float3(0, 0, -1);

//Scalars
float particleOpacity;
float timeInterpolation;
bool doHeatHaze;
bool doShadowMapping;

//Scalars
float kA <
    string UIWidget = "slider";
    float UIMin = 0.0;
    float UIMax = 1.0;
    float UIStep = 0.05;
    string UIName =  "AmbientIntensity";
> = 0.3;

float kD <
    string UIWidget = "slider";
    float UIMin = 0.0;
    float UIMax = 1.0;
    float UIStep = 0.05;
    string UIName =  "DiffuseIntensity";
> = 0.3;

float kS <
    string UIWidget = "slider";
    float UIMin = 0.0;
    float UIMax = 1.0;
    float UIStep = 0.05;
    string UIName =  "SpecularIntensity";
> = 0.3;

float nS <
    string UIWidget = "slider";
    float UIMin = 1.0;
    float UIMax = 100.0;
    float UIStep = 1.0;
    string UIName =  "SpecularCoefficent";
> = 20;

//Textures
Texture2D opacityTexture;
Texture2D normalTexture;
Texture2D specularTexture;
Texture2D shadowMapTexture;
Texture2D bloomTexture;
Texture2D finalSceneTexture;
Texture2D heatHazeTexture;
Texture2D motionBlurDepthTexture;
Texture2D blurMaskTexture;

//Samplers
SamplerState samplerLinear
{
	Filter = MIN_MAG_MIP_LINEAR;
};

SamplerState samplerShadow
{
	Filter = MIN_MAG_MIP_LINEAR;
	AddressU = Clamp;
	AddressV = Clamp;
};

SamplerState samplerSkybox
{
	Filter = MIN_MAG_MIP_LINEAR;
	AddressU = Clamp;
	AddressV = Clamp;
};


//_______ STRUCTS _______

struct VSIn
{
	float3 position : POSITION;
	float3 normal 	: NORMAL;
	float2 tex		: TEXCOORD;
};

struct PSIn
{
	float4 position : SV_POSITION;
	float3 normal 	: NORMAL;
	float2 tex		: TEXCOORD0;
	float3 light	: TEXCOORD1;
};

struct PSInMask
{
	float4 position : SV_POSITION;
	float color 	: COLOR;
};

struct PSInRenderTarget
{
	float4 position : SV_POSITION;
	float2 tex		: TEXCOORD0;
};

struct PSInShadow
{
	float4 position : SV_POSITION;
	float3 normal 	: NORMAL;
	float2 tex		: TEXCOORD0;
	float3 light	: TEXCOORD1;
	float4 posLight	: TEXCOORD2;
	float3 viewVec 	: TEXCOORD3;
	float3 tangent  : TEXCOORD4;
	float3 binormal : TEXCOORD5;
};

struct PSInDepth
{
	float4 position : SV_POSITION;
	float depth 	: TEXCOORD0;
};


//_______ FUNCTIONS _______

float4 GetEnlightened(float4 color, float specularMap, float3 normal, float3 lightDir, float3 viewDir)
{
	float nDotL = dot(normal, lightDir);
	float specular = 0;
	
	if (nDotL > 0) {
		float3 h = normalize(viewDir + lightDir);
		specular = kS * specularMap * lightColorDiffuse * pow(saturate(dot(normal, h)), nS);
	}
	
	return float4(saturate((kA * lightColorAmbient + 
					 		kD * lightColorDiffuse * saturate(nDotL) + 
					 		specular) * color.rgb), color.a);
}


//_______ SHADERS _______


//------- Standard

PSIn VSStandard(VSIn IN) 
{
	PSIn OUT = (PSIn)0;
	
	OUT.position = mul(float4(IN.position, 1), mWorldViewProj);
	OUT.normal = mul(float4(IN.normal, 0), mWorldView).xyz;
	OUT.light = mul(float4(lightDirection, 0), mView).xyz;
	OUT.tex = IN.tex;
	
	return OUT;
}

float4 PSStandard(PSIn IN) : SV_Target
{	
	return GetEnlightened(opacityTexture.Sample(samplerLinear, IN.tex), 
						  1, //specularTexture.Sample(samplerLinear, IN.tex).r, 
						  normalize(IN.normal), 
						  normalize(IN.light), 
						  viewDirection);
}


//------- Masking

PSInMask VSMaskBlack(VSIn IN) 
{
	PSInMask OUT = (PSInMask)0;
	
	OUT.position = mul(float4(IN.position, 1), mWorldViewProj);
	OUT.color.r = 0;
	
	return OUT;
}

PSInMask VSMaskWhite(VSIn IN) 
{
	PSInMask OUT = (PSInMask)0;
	
	OUT.position = mul(float4(IN.position, 1), mWorldViewProj);
	OUT.color.r = 1;
	
	return OUT;
}

float4 PSMask(PSInMask IN) : SV_Target
{	
	return float4(IN.color, 0, 0, 1);
}


//------- Skybox

PSIn VSSkybox(VSIn IN) 
{
	PSIn OUT = (PSIn)0;
	
	OUT.position = mul(float4(IN.position, 1), mWorldViewProj);
	float2 tex = IN.tex;
	if (tex.x == 0)
		tex.x += .001;
	if (tex.x == 1)
		tex.x -= .001;
	if (tex.y == 0)
		tex.y += .001;
	if (tex.y == 1)
		tex.y -= .001;
	
	OUT.tex = IN.tex;
	
	return OUT;
}

float4 PSSkybox(PSIn IN) : SV_Target
{	
	return opacityTexture.Sample(samplerSkybox, IN.tex);
}


//------- Depth map

PSInDepth VSDepthLight(VSIn IN) 
{
	PSInDepth OUT = (PSInDepth)0;
	
	OUT.position = mul(float4(IN.position, 1), mWorldViewProj);
	OUT.depth.r = 1 - OUT.position.z;
	
	return OUT;
}

PSInDepth VSDepth(VSIn IN) 
{
	PSInDepth OUT = (PSInDepth)0;
	
	OUT.position = mul(float4(IN.position, 1), mWorldViewProj);
	OUT.depth.r = OUT.position.z / 2500 *.5 + .5; 	//2300.0 = Farplane bei persp. Proj.
	
	return OUT;
}

float4 PSDepth(PSInDepth IN) : SV_Target
{
	return float4( IN.depth.r, 0, 0, 1 );
}


//------- Shadow mapping

PSInShadow VSShadow(VSIn IN) 
{
	PSInShadow OUT = (PSInShadow)0;
	
	OUT.position = mul(float4(IN.position, 1), mWorldViewProj);
	OUT.normal = mul(float4(IN.normal, 0), mWorldView).xyz;
	OUT.light = mul(float4(lightDirection, 0), mView).xyz;
	OUT.tex = IN.tex;
	OUT.viewVec = viewDirection;
	OUT.posLight = mul(float4(IN.position, 1), mul(mul(mWorld, mLight), mProjLight));
	
	return OUT;
}

float4 PSShadow(PSInShadow IN) : SV_Target
{
	float4 OUT = GetEnlightened(opacityTexture.Sample(samplerLinear, IN.tex), 
								1, //specularTexture.Sample(samplerLinear, IN.tex),
								normalize(IN.normal), 
								normalize(IN.light), 
								normalize(IN.viewVec));
	
	float3 shadowMap = IN.posLight.xyz / IN.posLight.w;
	float2 shadowTexCoords = 0.5 * shadowMap.xy + float2(0.5, 0.5);
	
	shadowTexCoords.y = 1.0f - shadowTexCoords.y;

	//Tiefe des aus der Lichtperspektive sichtbaren Pixels.
	float shadowMapDepth = shadowMapTexture.Sample(samplerShadow, shadowTexCoords).r;    

	//Tiefe des nun zu rendernden Pixels - aus Lichtperspektive.
	float currentDepth = 1 - shadowMap.z;

	//Bias um floating-point Fehler zu kompensieren.
	if (shadowMapDepth > 0 && shadowMapDepth - 0.005 > currentDepth)
		OUT.rgb *= .3;
	
	return OUT;
}


//------- Heat haze

float4 PSHeat(PSIn IN) : SV_Target
{
	float4 OUT = opacityTexture.Sample(samplerLinear, IN.tex);
	OUT.a *= particleOpacity;
	return OUT;
}


//------- Finalize plus Motionblur output

float4 PSFinalizeMotionBlur(PSInRenderTarget IN) : SV_Target
{
	float4 OUT = float4(0, 0, 0, 1);
	float4 hazeColor = heatHazeTexture.Sample(samplerLinear, IN.tex);
	float3 hazeNormal = hazeColor.rgb * 2 - 1;
	float2 hazeTexCoords = IN.tex;
	
	if (doHeatHaze)
		hazeTexCoords += hazeColor.a * 0.02 * hazeNormal.xy;
	
	//float blurMask = blurMaskTexture.Sample(samplerShadow, IN.tex).r;
		
	//Tiefe aus View-Perspektive.
	float Z = motionBlurDepthTexture.Sample(samplerShadow, IN.tex).r;

	//Aktuelle Viewport-Position des Pixels (-1 bis +1).
	float4 currentPos = float4(IN.tex.x * 2.0f - 1.0f, IN.tex.y * 2.0f - 1.0f, (Z * 2.0f - 1.0f) * 2500.0f, 1);
	
	float4 D = mul(currentPos, mViewProjInverse);	//Ursprungsposition herstellen.
	float4 worldPos = D / D.w;						//Inhomogen'isieren' für worldPos.
	float4 previousPos = mul(worldPos, mViewProjPrevious); //Position im vorigen Frame.
	//float W = previousPos.w;
	previousPos /= previousPos.w;					//Inhomogen'isieren'.
	
	//Differenz der beiden = Geschwindigkeit.
	float3 velocity = (currentPos - previousPos).xyz;// / -1500.0f;// timeInterpolation;
	//velocity *= W;
	velocity /= velocity.z;
	velocity *= .2f;
	
	int sampleAnz = 20; //blurMask == 1 ? 20 : 1;
	
	for(int i = 0; i < sampleAnz; i++, hazeTexCoords += velocity) {
		 OUT.rgb += finalSceneTexture.Sample(samplerShadow, hazeTexCoords).rgb;
	}
	
	OUT.rgb /= sampleAnz;
	
	return OUT;
}


float4 PSFinalize(PSInRenderTarget IN) : SV_Target
{
	float4 OUT = 1;
	float4 hazeColor = heatHazeTexture.Sample(samplerLinear, IN.tex);
	float3 hazeNormal = hazeColor.rgb * 2.0 - 1.0;
	float2 hazeTexCoords = IN.tex;
	
	if (doHeatHaze)
		hazeTexCoords += hazeColor.a * 0.02 * hazeNormal.xy / hazeNormal.z;
		
	OUT.rgb = finalSceneTexture.Sample(samplerLinear, hazeTexCoords.xy);
	
	return OUT;
}


//------- Rendertarget output

PSInRenderTarget VSRenderTarget(VSIn IN) 
{
	PSInRenderTarget OUT = (PSInRenderTarget)0;
	
	OUT.position = float4(IN.position, 1);
	OUT.tex = IN.tex;
	
	return OUT;
}

float4 PSRenderTarget(PSInRenderTarget IN) : SV_Target
{
	return float4(opacityTexture.Sample(samplerLinear, IN.tex).rgb, 1);
}


//_______ STATES _______

RasterizerState DisableCulling { 
	CullMode = NONE;
	FillMode = SOLID;
};
RasterizerState CullBack { 
	CullMode = BACK;
	FillMode = SOLID;
	FrontCounterClockwise = true;
};
RasterizerState CullFront { 
	CullMode = FRONT;
	FillMode = SOLID;
	FrontCounterClockwise = true;
};
DepthStencilState EnableDepth { 
	DepthEnable = TRUE;
	DepthWriteMask = ALL;
	DepthFunc = LESS;
};
DepthStencilState DisableDepth {
	DepthEnable = FALSE;
	DepthWriteMask = ZERO;
};
BlendState DisableBlend { 
	BlendEnable[0] = FALSE; 
};
BlendState EnableBlend { 
	BlendEnable[0] = TRUE;
	SrcBlend = SRC_ALPHA;
	DestBlend = INV_SRC_ALPHA;
	BlendOp = ADD;
};


//_______ TECHNIQUES _______

technique10 standard
{
	pass P0
	{
		SetGeometryShader( 0 );
		SetVertexShader( CompileShader( vs_4_0, VSStandard() ) );
		SetPixelShader( CompileShader( ps_4_0, PSStandard() ) );
		
		//SetRasterizerState(CullBack);
	}
};

technique10 maskBlack
{
	pass P0
	{
		SetGeometryShader( 0 );
		SetVertexShader( CompileShader( vs_4_0, VSMaskBlack() ) );
		SetPixelShader( CompileShader( ps_4_0, PSMask() ) );
		
		//SetRasterizerState(CullBack);
	}
};

technique10 maskWhite
{
	pass P0
	{
		SetGeometryShader( 0 );
		SetVertexShader( CompileShader( vs_4_0, VSMaskWhite() ) );
		SetPixelShader( CompileShader( ps_4_0, PSMask() ) );
		
		//SetRasterizerState(CullBack);
	}
};

technique10 skybox
{
	pass P0
	{
		SetGeometryShader( 0 );
		SetVertexShader( CompileShader( vs_4_0, VSSkybox() ) );
		SetPixelShader( CompileShader( ps_4_0, PSSkybox() ) );
		
		//SetRasterizerState(CullFront);
	}
};

technique10 depth
{
	pass P0
	{
		SetGeometryShader( 0 );
		SetVertexShader( CompileShader( vs_4_0, VSDepth() ) );
		SetPixelShader( CompileShader( ps_4_0, PSDepth() ) );
		
		//SetRasterizerState(CullBack);
	}
};

technique10 depthLight
{
	pass P0
	{
		SetGeometryShader( 0 );
		SetVertexShader( CompileShader( vs_4_0, VSDepthLight() ) );
		SetPixelShader( CompileShader( ps_4_0, PSDepth() ) );
		
		//SetRasterizerState(CullBack);
	}
};

technique10 shadow
{
	pass P0
	{
		SetGeometryShader( 0 );
		SetVertexShader( CompileShader( vs_4_0, VSShadow() ) );
		SetPixelShader( CompileShader( ps_4_0, PSShadow() ) );
		
		//SetRasterizerState(CullBack);
	}
};

technique10 heat
{
	pass P0
	{
		SetGeometryShader( 0 );
		SetVertexShader( CompileShader( vs_4_0, VSStandard() ) );
		SetPixelShader( CompileShader( ps_4_0, PSHeat() ) );
		
		//SetRasterizerState(CullBack);
		SetBlendState(EnableBlend, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xffffff);
	}
};

technique10 finalize
{
	pass P0
	{
		SetGeometryShader( 0 );
		SetVertexShader( CompileShader( vs_4_0, VSRenderTarget() ) );
		SetPixelShader( CompileShader( ps_4_0, PSFinalize() ) );
		
		//SetRasterizerState(CullBack);
		//SetBlendState(DisableBlend, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
	}
};

technique10 finalizeMotionBlur
{
	pass P0
	{
		SetGeometryShader( 0 );
		SetVertexShader( CompileShader( vs_4_0, VSRenderTarget() ) );
		SetPixelShader( CompileShader( ps_4_0, PSFinalizeMotionBlur() ) );
		
		//SetRasterizerState(CullBack);
		//SetBlendState(DisableBlend, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
	}
};

technique10 rendertarget
{
	pass P0
	{
		SetGeometryShader( 0 );
		SetVertexShader( CompileShader( vs_4_0, VSRenderTarget() ) );
		SetPixelShader( CompileShader( ps_4_0, PSRenderTarget() ) );
		
		//SetRasterizerState(CullBack);
		//SetBlendState(DisableBlend, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF);
	}
};